#!/usr/bin/env sage

n = 64

################################################################

import struct, hashlib

def unpack_u(r):
    assert len(r) == n*26//8
    assert n % 4 == 0
    u = [0] * n
    for i in range(n//4):
        u[4*i+0] |= r[13*i+ 0] <<  0
        u[4*i+0] |= r[13*i+ 1] <<  8
        u[4*i+0] |= r[13*i+ 2] << 16
        u[4*i+0] |= (r[13*i+ 3] & ((1 << 2) - 1)) << 24
        u[4*i+1] |= r[13*i+ 3] >>  2
        u[4*i+1] |= r[13*i+ 4] <<  6
        u[4*i+1] |= r[13*i+ 5] << 14
        u[4*i+1] |= (r[13*i+ 6] & ((1 << 4) - 1)) << 22
        u[4*i+2] |= r[13*i+ 6] >>  4
        u[4*i+2] |= r[13*i+ 7] <<  4
        u[4*i+2] |= r[13*i+ 8] << 12
        u[4*i+2] |= (r[13*i+ 9] & ((1 << 6) - 1)) << 20
        u[4*i+3] |= r[13*i+ 9] >>  6
        u[4*i+3] |= r[13*i+10] <<  2
        u[4*i+3] |= r[13*i+11] <<  10
        u[4*i+3] |= r[13*i+12] <<  18
    return vector(u)

def pack_u(u):
    assert len(u) == n
    r = [0] * (n*26//8)
    for i in range(n//8):
        r[26*i + 0] |= 0xff & (u[(i<<3)])
        r[26*i + 1] |= 0xff & (u[(i<<3)]>>8)
        r[26*i + 2] |= 0xff & (u[(i<<3)]>>16)
        r[26*i + 3] |= 0xff & ((u[(i<<3)]>>24)|(u[(i<<3)+1]<<2))
        r[26*i + 4] |= 0xff & (u[(i<<3)+1]>>6)
        r[26*i + 5] |= 0xff & (u[(i<<3)+1]>>14)
        r[26*i + 6] |= 0xff & ((u[(i<<3)+1]>>22)|(u[(i<<3)+2]<<4))
        r[26*i + 7] |= 0xff & (u[(i<<3)+2]>>4)
        r[26*i + 8] |= 0xff & (u[(i<<3)+2]>>12)
        r[26*i + 9] |= 0xff & ((u[(i<<3)+2]>>20)|(u[(i<<3)+3]<<6))
        r[26*i +10] |= 0xff & (u[(i<<3)+3]>>2)
        r[26*i +11] |= 0xff & (u[(i<<3)+3]>>10)
        r[26*i +12] |= 0xff & (u[(i<<3)+3]>>18)
        r[26*i +13] |= 0xff & (u[(i<<3)+4])
        r[26*i +14] |= 0xff & (u[(i<<3)+4]>>8)
        r[26*i +15] |= 0xff & (u[(i<<3)+4]>>16)
        r[26*i +16] |= 0xff & ((u[(i<<3)+4]>>24)|(u[(i<<3)+5]<<2))
        r[26*i +17] |= 0xff & (u[(i<<3)+5]>>6)
        r[26*i +18] |= 0xff & (u[(i<<3)+5]>>14)
        r[26*i +19] |= 0xff & ((u[(i<<3)+5]>>22)|(u[(i<<3)+6]<<4))
        r[26*i +20] |= 0xff & (u[(i<<3)+6]>>4)
        r[26*i +21] |= 0xff & (u[(i<<3)+6]>>12)
        r[26*i +22] |= 0xff & ((u[(i<<3)+6]>>20)|(u[(i<<3)+7]<<6))
        r[26*i +23] |= 0xff & (u[(i<<3)+7]>>2)
        r[26*i +24] |= 0xff & (u[(i<<3)+7]>>10)
        r[26*i +25] |= 0xff & (u[(i<<3)+7]>>18)
    return bytes(r)

################################################################

def encode_pk(pk):
    h1,h2 = pk
    return pack_u(h1) + pack_u(h2)

def encode_sk(sk, pk):
    x1,x2,F1,F2 = sk
    h1,h2 = pk
    return b''.join([
            struct.pack(f'<{n}b', *x1),
            struct.pack(f'<{n}b', *x2),
            struct.pack(f'<{n}b', *F1[0]),
            struct.pack(f'<{n}i', *F1[1]),
            struct.pack(f'<{n}b', *F2[0]),
            struct.pack(f'<{n}i', *F2[1]),
            hashlib.sha3_256(encode_pk(pk)).digest(),
        ])

def decode_pk(pk):
    h1 = vector(unpack_u(pk[:len(pk)//2]))
    h2 = vector(unpack_u(pk[len(pk)//2:]))
    res = (h1,h2)
    assert pk == encode_pk(res)
    return res

def decode_sk(sk):
    x1 = vector(struct.unpack(f'<{n}b', sk[0*n:1*n]))
    x2 = vector(struct.unpack(f'<{n}b', sk[1*n:2*n]))
    F1_0 = vector(struct.unpack(f'<{n}b', sk[2*n:3*n]))
    F1_1 = vector(struct.unpack(f'<{n}i', sk[3*n:7*n]))
    F2_0 = vector(struct.unpack(f'<{n}b', sk[7*n:8*n]))
    F2_1 = vector(struct.unpack(f'<{n}i', sk[8*n:12*n]))
    pkh = sk[12*n:]
    res = (x1,x2, (F1_0,F1_1), (F2_0,F2_1))
    assert sk[:12*n] == encode_sk(res, ([0]*64,)*2)[:12*n]
    return res

